/*$Id: lsn.c,v 1.13 2001/06/01 15:48:53 yuk Exp $

   lsn -     
   Copyright (C) 1998-2000, <Yura Kalinichenko> yuk@kron.vinnica.ua

     :

	lsn [] 1st_port 2nd_port ....

    :

       1st_port -   , . /dev/tty1a
       2nd_port -   , . /dev/tty2a
       ....     -   -   
    :
       -q -       
            (   )
       -h -  .  - 
       -t min  -    .  - 1 .
       -v -      .
       -n - - .  - .
       -s baud -  .  - 38400.
       -S baud - formal baud rate
       -b size -   .  - 1500 .
       -l logfile -   .  - stderr
----------------------------------------------------------------
$Log: lsn.c,v $
Revision 1.13  2001/06/01 15:48:53  yuk
bugfix: LOGHD=stderr

Revision 1.12  2001/05/31 07:53:32  yuri
Replace #include <ioctls.h> to <sys/ioctl.h>

Revision 1.11  2001/05/28 15:28:50  yuk
   

Revision 1.10  2000/11/22 07:56:10  yuk
added info lines: duplex, flow control

Revision 1.9  2000/11/13 09:54:50  yuk
new #include for Linux kernel 2.2.x

Revision 1.8  2000/11/11 10:54:14  yuk
FreeBSD newanses: pause before signal for wakeuping of sons.

Revision 1.7  2000/11/11 10:19:52  yuk
SCO UnixWare 7 newanses:
- can't close ports in processes (???)
- son started too early - we must make pause before father write to port
  the first time

Revision 1.6  2000/11/10 15:46:30  yuk
bugfix: in halfduplex mode read the rest of damaged block from right port

Revision 1.5  2000/11/10 14:26:42  yuk
   (- , )  Linux

Revision 1.4  2000/11/10 11:08:03  yuk
  

Revision 1.3  2000/11/10 09:11:43  yuk
-       ( 2  ).
   -      .
-     .

Revision 1.2  2000/11/10 07:43:06  yuk
don't close processes until done write message.

Revision 1.1  2000/11/09 08:11:02  yuk
 ,  select,   .  :
- Linux
- FreeBSD
- SCO UnixWare 7.1

Revision 1.0  1998/12/03 13:46:05  yuk
 
----------------------------------------------------------------
*/
#ifdef BSD
#define SETSPEED 1
#endif

#ifdef SCO
#define SETSPEEDIO 1
#endif

#define MAX_PROC 64

#include <stdio.h>
#include <termios.h>
#include <unistd.h>
#include <stdlib.h>
#include <fcntl.h>
#include <sys/ioctl.h>
#include <signal.h>
#include <locale.h>
#include <time.h>
#include <sys/wait.h>
#include <sys/timeb.h>
#include <sys/utsname.h>

#include <sys/times.h>
#include <sys/types.h>

#ifdef SCO
#include	<sys/termiox.h>
#endif

#ifdef CPULOAD
#include <proc/readproc.h>
#define MIN_TIME 1.0
#endif

#define START_CHAR '\1'
#define END_CHAR '\2'

#ifndef CRTSCTS
#ifdef RTSFLOW
#define CRTSCTS RTSFLOW
#else
#ifdef RTSXOFF
#define CRTSCTS (RTSXOFF | CTSXON)
#else
#error "Don't know how to set hardware flow control"
#endif
#endif
#endif

#define TRUE 1
#define FALSE 0
#define Tout 10L
#define elen 5

/*    - */
#define BLOCK_S   '\0'
#define TIME_S	  '\10'

/*    / */

extern void prtsig();

extern int optind, opterr, optopt;
extern char *optarg;

struct termios old_trmpar_1, new_trmpar_1;
struct termios old_trmpar_2, new_trmpar_2;
int done = FALSE;
int start = FALSE;
fd_set rfds;
char rot;
char usage[] = { "lsn $Revision: 1.13 $\n\
Usage:\n\t%s [options] 1st_line 2nd_line ....\n\n\
Options:\n\
\t\t-q	       - No output to stdout \n\
\t\t-n         - Null modem line. Default - full modem\n\
\t\t-h         - Half duplex. Default - full duplex\n\
\t\t-l logfile - Write statistics into logfile. Default - stderr\n\
\t\t-t min     - Test duration in minutes. Default - 1\n\
\t\t-s baud    - Real Baud rate. Default - 38400.\n\
\t\t-S baud    - Baud rate for termios. Default = Real Baud rate.\n\
\t\t-b size    - Block size. Default - 1500 bytes.\n\n" };

int Speed, null, lbuf, verbose, quiet, duplex;
char *logfile = NULL;
time_t test_duration;
int rows = 0;
int columns = 0;
FILE *LOGHD;

#define ERR_MEM 1
#define ERR_READ  2
#define ERR_WRITE 3
#define ERR_SELECT 4
#define ERR_CMP 5
#define ERR_LEN 6

#define status_len 25
int show_status(int n, int ok, int bad) {
int x, y, l;

x=status_len * (n/rows) + 1;
y=(n)%rows + 1;
printf("\033[%d;%dHProcess %#02d: %#05d/%#03d\r",
  y, x, n, ok, bad);
/*
printf("\033[%d;%dH%#02d;%#02dProcess %#02d: %#05d/%#03d\r",
  y, x, y, x, n, ok, bad);
*/
fflush(stdout);
}

int transfer(int procn, int fromhd, int tohd, int blsize, 
	     int* blocks, int* errors) {
  pid_t son;
  int s2pipe[2];
  int status, i, j;
  char *inbuf, *outbuf, *p;
  char msgbuf[1];
  int  pass = 0;

     *blocks = *errors = 0;  
     if ((inbuf=(char*) calloc(blsize, 1)) == NULL) return(ERR_MEM);
     inbuf[0] = START_CHAR;
     for (i=1; i<blsize-1; i++) inbuf[i] = '0' + (0177 & i);
     inbuf[blsize-1] = END_CHAR;
     if (pipe(s2pipe) < 0) {
       perror("level 2 pipe");
       exit(1);
     }
     
#ifdef DEBUG
    fprintf(stderr, "Start transfer\n", i);
#endif

     /*    :     1,
				           2 */
     if ((son = fork()) == 0) {
        close(s2pipe[0]);
        if ((outbuf=(char*) calloc(blsize, 1)) == NULL) exit(ERR_MEM);
	/*
	 *    -      2
	 */
	while (!done) { /*     */
	   j = 0;
	   p = outbuf;
           /*     outbuf */
#ifdef DEBUG
//    fprintf(stderr, "read start\n", i);
#endif
start_read:
           while (!done && 
             ((i=read(pass ? tohd : fromhd, p, blsize-j)) > 0)) { /*    */
	      j += i;
#ifdef DEBUG
//    fprintf(stderr, "read %d:%d chars\n", i, j);
#endif
	      if (j >= blsize) break;
              p = &p[i];
           }
#ifdef DEBUG
//    fprintf(stderr, "read end\n");
#endif
           if (i < 0) status=ERR_READ;
           else if (j != blsize) status = ERR_LEN;
                else  /*      	*/
                  status = (memcmp(inbuf, outbuf, blsize) == 0) ? 0 : ERR_CMP;

           /*    	*/
           outbuf[0] = status;
           if (!done & status) { /*   -  */
              /*     	*/
              write(s2pipe[1], outbuf, 1);
              /*       */
              while (!done && (i=read(pass ? tohd : fromhd, outbuf, 1)) > 0)
                if (outbuf[0] == START_CHAR) {
//fprintf(stderr, "error: %0x\n", outbuf[0]);
                   j = 1;
                   p = &outbuf[1];
                   goto start_read;
                }
           }

           if (!duplex) pass = ~pass;
           /*     	*/
           outbuf[0] = '\0';
           write(s2pipe[1], outbuf, 1);
        }
#ifdef DEBUG
    fprintf(stderr, "son died\n");
#endif
        exit(0);
     } else {         /* father	*/
        close(s2pipe[1]);
        if (duplex) {
          i = fcntl(s2pipe[0], F_GETFL);
          fcntl(s2pipe[0], F_SETFL, i | O_NDELAY);
        }
        j = 0;
#ifdef SCO
        usleep(50000L);
#endif
        while (!done) { /* while not signals */
#ifdef DEBUG
//    fprintf(stderr, "write start\n", i);
#endif
           write(pass ? fromhd : tohd, inbuf, blsize);	/* send block to port */
           (*blocks)++;				/* block counter      */
#ifdef DEBUG
//    fprintf(stderr, "write end\n", i);
#endif
           if (!done) { 
              while ((i=read(s2pipe[0], msgbuf, 1)) > 0) {
                    if (msgbuf[0]) (*errors)++;
                    else if (!duplex) pass = ~pass;
                    if (!duplex) break;
                    else i--;
              }
              if (!quiet) show_status(procn, *blocks, *errors);
           }
           else { 
              while ((i=read(s2pipe[0], msgbuf, 1)) > 0);
           }
        }
#ifdef DEBUG
    fprintf(stderr, "all writes ended\n");
#endif
        kill(son, SIGHUP);
        /*      */
#ifdef DEBUG
    fprintf(stderr, "wait for son\n");
#endif
        waitpid(son, &status, 0);
#ifdef DEBUG
    fprintf(stderr, "wait ended\n");
#endif
     }
     return(status);
}

int np = 0;
struct process_t {
   char *dev1;
   char *dev2;
   int id;
   int spipe[2];
} family[MAX_PROC];

main(argc, argv)
int argc;
char **argv;
{
  char *obuf1;
  int i, j, k;
  int error_count;
  int status;
  float f;
  int nspeed, sspeed;
  double counter;
  time_t start_time, end_time;
  int blocks = 0;
  int errors = 0;
  char buf[80];
  struct tm *loct;
  time_t ct;
  struct utsname os;
  int hd1, hd2;
  struct winsize ws;
#ifdef CPULOAD
  int cpuavg;
#endif
  
  setlocale(LC_ALL, "");

  error_count = 0;
  if (argc < 3) { /*   :  < 3  -  */
     error_count++;
  }
  else {
        /*
         *    
         */

        /* Defaults */
        Speed = B38400;
        nspeed = 38400;
        sspeed = 0;
        null = FALSE;
        duplex = TRUE;
        lbuf = 1500;
        test_duration = 1;	/* minutes */
        verbose = FALSE;
        quiet = FALSE;
        if (i = ioctl (fileno(stdout), TIOCGWINSZ, (char *) &ws)) {
           rows = 0;
           columns = 0;
        }
        else {
           rows = ws.ws_row-1;
           columns = ws.ws_col;
        }
        if (!rows) rows = 24;
        if (!columns) columns = 80;

        /*      */
        while ((i = getopt(argc, argv, "qdnHh?l:t:S:s:b:")) != EOF) {
          switch (i) {

          case 'n': /* stty clocal */
                    null = TRUE;
                    break;
          
          case 'q': /* quiet output */
                    quiet = TRUE;
                    break;

          case 'h': /* half duplex */
                    duplex = FALSE;
                    break;
          
          case 't': /* test duration in minutes */
                    test_duration = atoi(optarg);
                    break;
                    
          case 'b': /*   */
                    lbuf = atoi(optarg);
                    if (lbuf == 0) {
                       fprintf(stderr, "Block size must be greater then 0\n");
                       exit(1);
                    }
                    break;
          case 'l': /*  logfile 	*/
          	    logfile = optarg;
          	    break;
          case 's': /*    */
                    nspeed = atoi(optarg);
                    break;
          case 'S': /*    */
                    sspeed = atoi(optarg);
                    break;

          case 'H':
          case '?':
                    error_count++;
                    break;
                    
          default:  /*    -                    */
                    fprintf(stderr, "Unknown key: -%c\n", i);
                    error_count++;
                    break;
          }
        }
  }
  
  if (!sspeed) sspeed = nspeed;
  switch (sspeed) {
  case 50:
  	     Speed = B50;
  	     break;
  case 75:
  	     Speed = B75;
  	     break;
  case 134:
  	     Speed = B134;
  	     break;
  case 150:
  	     Speed = B150;
  	     break;
  case 200:
  	     Speed = B200;
  	     break;
  case 600:
  	     Speed = B600;
  	     break;
  case 1200:
  	     Speed = B1200;
  	     break;
  case 1800:
  	     Speed = B1800;
  	     break;

  case 2400: Speed = B2400;
             break;
  case 4800: Speed = B4800;
             break;
  case 9600: Speed = B9600;
             break;
#ifdef B19200
  case 19200:
	     Speed = B19200;
	     break;
#else
#ifdef EXTA				
  case 19200:
	     Speed = EXTA;
	     break;
#endif
#endif

#ifdef B38400
  case 38400:
	     Speed = B38400;
	     break;
#else
#ifdef EXTB				
  case 38400:
	     Speed = EXTB;
	     break;
#endif
#endif

#ifdef B57600
  case 57600: 
             Speed = B57600;
             break;
#endif
#ifdef B115200
  case 115200: 
             Speed = B115200;
             break;
#endif
   default:  
             error_count++;
	     fprintf(stderr, "Illegal line speed: %d\n", nspeed);
             break;
   }

        if (error_count) {                 /*       */
           fprintf(stderr, usage, argv[0]);    /*               */
           exit(1);                   /*                */
        }

  /*      prtsig */
  for (i=1; i<NSIG; i++) signal(i, prtsig);

  np = 0;
  while (optind < argc) {
    if ((np+1) >= MAX_PROC) {
       fprintf(stderr, "ERROR: Too many lines.\n");
       exit(1);
    }
    family[np].dev1 = argv[optind++];
    family[np++].dev2 = argv[optind++];

    if (duplex) {    
       family[np].dev1 = family[np-1].dev2;
       family[np++].dev2 = family[np-1].dev1;
    }
  }

  /* ,     */
  for (i=0; i<np; i++) {  
    /*    */
    if ((hd1 = open(family[i].dev1, O_RDWR )) == -1) { 
       perror(family[i].dev1);
       exit(1);
    }
    /*    */
    if ((hd2 = open(family[i].dev2, O_RDWR )) == -1) {
       perror(family[i].dev2);
       exit(1);
    }
    close(hd1);
    close(hd2);
    pipe(family[i].spipe);	/*      */
  }
    
  /*   (ANSI X3.64) */
  if (!quiet) printf("\033[2J\n");

  for (i=0; i<np; i++) {	/* create all sons	*/
#ifdef DEBUG
    fprintf(stderr, "Open transfer session %d\n", i);
#endif
    if ((family[i].id=fork()) == 0) {	/*  i:	hd1 -> hd2 */
        close(family[i].spipe[0]);
        if ((hd1 = open(family[i].dev1, O_RDWR )) == -1) { 
           perror(family[i].dev1);
           errors++;
        }
        /*     -    */
        else if ((hd2 = open(family[i].dev2, O_RDWR )) == -1) {
                perror(family[i].dev2);
                errors++;
        }
        /*     -   */
        else { 
          init_line(hd1, &old_trmpar_1); /*   1 */
          init_line(hd2, &old_trmpar_2); /*   2 */
        }
        buf[0] = errors;
        write(family[i].spipe[1], buf, 1);
        
        if (!errors) {
           pause(); /* wait for start */
           transfer(i, hd1, hd2, lbuf, &blocks, &errors);
           /*     */
           sprintf(buf, "%05d:%05d", blocks, errors);
           /*  */
           write(family[i].spipe[1], buf, 11);
#ifndef SCO
           close(hd1);
           close(hd2);
#endif
           sleep(1);
        }
        exit(errors);
    }
  }
  for (i=0; i<np; i++) {
     close(family[i].spipe[1]);
  }
#ifdef CPULOAD
  cpuavg = get_cpuload(MIN_TIME);
//  printf("CPU 1 = %d\n", cpuavg);
#endif
#ifdef DEBUG
  fprintf(stderr, "np=%d\n", np);
#endif
  for (i=0; i<np; i++) {
     if ((j=read(family[i].spipe[0], buf, 1)) > 0) {
#ifdef DEBUG
	fprintf(stderr, "waiting for pipe %d\n", family[i].id);
#endif     
     ;
//fprintf(stderr, "read for process %d\n", family[i].id);
        errors += buf[0];
     }
     else {
       errors++;
       perror("read from pipe");
     }
  }

  if (errors) {
     for (i=0; i<np ; i++)
        if (family[i].id) kill(family[i].id, SIGKILL);
     exit(1);
  }
  else {
     time(&start_time);
     loct = localtime(&start_time);
#ifdef BSD
     usleep(10);
#endif     
     alarm(test_duration*60);
     for (i=0; i<np ; i++)
        if (family[i].id) kill(family[i].id, SIGUSR1);
  }  

  while (!done) pause();
  
  time(&end_time);

  for (i=0; i<np; i++) {
//fprintf(stderr, "Wait for process %d\n", family[i].id);
//     waitpid(family[i].id, &status, 0);
//fprintf(stderr, "End of process %d with status %d\n", 
//        family[i].id, WEXITSTATUS(status));
     if ((j=read(family[i].spipe[0], buf, 11)) > 0) {
#ifdef DEBUG
	fprintf(stderr, "waiting for pipe %d\n", family[i].id);
#endif     
     ;
//fprintf(stderr, "read for process %d\n", family[i].id);
        buf[j] = '\0';
        sscanf(buf,"%d:%d", &j, &k);
        blocks += j;
        errors += k;
     }
     else perror("read from pipe");
  }

  for (i=0; i<np; i++) {
//fprintf(stderr, "Wait for process %d\n", family[i].id);
     waitpid(family[i].id, &status, 0);
//fprintf(stderr, "End of process %d with status %d\n", 
//        family[i].id, WEXITSTATUS(status));
  }
#ifdef CPULOAD
  cpuavg = get_cpuload(MIN_TIME);
//  printf("CPU 2 = %d; Average = %d\n", cpu, cpuavg);
#endif
  /* statistics	*/
  LOGHD=stderr;
  if (logfile) {
//     fclose(stdout);
     if ((LOGHD = fopen(logfile, "w")) == NULL) {
        perror(logfile);
        LOGHD = stderr;
     }
  }
  
  fprintf(LOGHD, "Statistic on serial test.\n");
  fprintf(LOGHD, "=========================\n");
  strftime(buf, 80, "%H:%M:%S -- %d %B %Y", loct);
  fprintf(LOGHD, "Test started: %s\n", buf);
  loct = localtime(&end_time);
  strftime(buf, 80, "%H:%M:%S -- %d %B %Y", loct);
  fprintf(LOGHD, "Test ended:   %s\n", buf);
  fprintf(LOGHD, "=========================\n");
  uname(&os);
  fprintf(LOGHD, "OS %s rel.%s ver.%s at %s\n", 
  	os.sysname, os.release, os.version, os.machine);
  fprintf(LOGHD, "Number of ports: %d\n", np);
  fprintf(LOGHD, "%s duplex\n", duplex ? "Full" : "Half");
  fprintf(LOGHD, "Flow control:    %s\n", null ? "XON/XOFF" : "RTS/CTS");
  fprintf(LOGHD, "Speed at line:   %d\n", nspeed);
  fprintf(LOGHD, "Data block:      %d\n", lbuf);
  fprintf(LOGHD, "Test time (min): %8.3f\n", (end_time-start_time)/60.0);
#ifdef CPULOAD
  fprintf(LOGHD, "Average CPUload: %d.%d%%\n", cpuavg / 10, cpuavg % 10);
#endif
  fprintf(LOGHD, "Average CPS:     %d\n", ((blocks-errors)*lbuf)/np/(end_time-start_time));
  fprintf(LOGHD, "Errors:          %8.3f%%\n", (errors*100.0)/blocks);
  fprintf(LOGHD, "Blocks were transferred:\n\
\tTotal: %ld\n\
\tAvarage at port: %ld\n", blocks, blocks/np);
  fprintf(LOGHD, "Errors:\n\
\tTotal: %ld\n\
\tAvarage at port: %ld\n", errors, errors/np);
  exit(0);
}

init_line(fd, old_stty)
int fd;				/* file descriptor */
struct termios *old_stty;	/*     */
{
  struct termios stty;
  int i;
  
  tcgetattr(fd, old_stty); /*    */
  /*    a     */
  memcpy(&stty, old_stty, sizeof(stty));
  /*        ,  :    */
  stty.c_iflag = IGNBRK | IGNPAR | IGNCR ;
  stty.c_cflag = CS8 | CREAD | Speed ;
  stty.c_oflag = 0;
  stty.c_lflag = 0;
  stty.c_cc[VMIN] = BLOCK_S; /*   .  */
  stty.c_cc[VTIME] = TIME_S; /*   */
  if (null) {                /*  - :        */
    stty.c_iflag |= (IXON | IXOFF );     /*   xon/xoff */
    stty.c_cflag &= ~(CRTSCTS); /*   RTS/CTS */
    stty.c_cflag |= (CLOCAL);   /*     */
  }
  else {                    /*   :         */
    stty.c_iflag &= ~(IXON | IXANY | IXOFF ); /*  xon/xoff         */
    stty.c_cflag |= (CRTSCTS);       /*   RTS/CTS  */
    stty.c_cflag &= ~(CLOCAL);       /*     */
  }

  /*    stty- */
  
#ifdef SETSPEED
  cfsetspeed(&stty, Speed);
#endif
#ifdef SETSPEEDIO
  cfsetispeed(&stty, Speed);
  cfsetospeed(&stty, Speed);
#endif

  if (tcsetattr(fd, TCSANOW, &stty) < 0) {
     perror("IOCTL");
     exit(1);
  }
}

close_line(fd, old_stty)
int fd;
struct termios old_stty;	/*     */
{
   tcsetattr(fd, TCSANOW, &old_stty);
   close(fd);
}

/*
 *   
 */
void prtsig(num)
int num;
{
  int i;
  
#ifdef DEBUG
  fprintf(stderr, "\nWe are %d: ", getpid());
#endif
  /*     */
  switch (num) {
   case SIGUSR1 :
   		  start = 1;
   		  break;
   case SIGALRM	:
#ifdef DEBUG
  fprintf(stderr, "SIGALARM\n");
#endif
                  for (i=0; i<np ; i++)
                  if (family[i].id) { 
                   kill(family[i].id, SIGHUP);
//                   family[i].id = 0;
                  }
                  done = 1;
		  break;
   default:	
#ifdef DEBUG
  fprintf(stderr, "%d\n", num);
#endif
		  done = 1;	/*   !!! */
   		  break;
   }
   return ;
}
